# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb.

# %% auto 0
__all__ = ['RQBottleneckTransformer', 'RQVisual', 'make_model']

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 2
import io
import time
import torch
import torchaudio

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 3
from pathlib import Path
import json
from fastprogress import progress_bar, master_bar
import fastprogress
import numpy as np
import pylab as plt
import pandas as pd
import random
import IPython

import whisper
from huggingface_hub import hf_hub_download
from fastcore.basics import store_attr

from torch import nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data.dataloader import DataLoader
import webdataset as wds
from . import vad, wh_transcribe

from vector_quantize_pytorch import ResidualVQ

from fastcore.script import *

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 7
import itertools
import math

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 8
class SADataset(torch.utils.data.IterableDataset):
    def __init__(self, basedir, data, tokenizer, bs=8, buffer_count=16, sample_rate=16000, audio_seconds=30, ttoks_size=200, random_shift=False):
        self.basedir = basedir
        self.data = data
        self.tokenizer = tokenizer
        self.bs = bs
        self.buffer_count = buffer_count
        self.files = pd.Series(data.groupby('fname').size().index)
        self.sample_rate = sample_rate
        self.audio_seconds = audio_seconds
        self.ttoks_size = ttoks_size
        self.random_shift = random_shift
        self._sr_cache = None

    def load_audio(self, fname):
        """Load an audio file to the GPU and resample to `self.sample_rate`."""
        x, sr = torchaudio.load(self.basedir/fname)
        if sr != self._sr_cache:
            self._sr_tform = torchaudio.transforms.Resample(sr, self.sample_rate)
            self._sr_cache = sr
        pad_size = self.audio_seconds * self.sample_rate
        return self._sr_tform(x)
        
    def __len__(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            n = len(self.data) / self.bs
        else:
            n = math.ceil(len(self.data) / worker_info.num_workers) / self.bs
        return math.ceil(n)
        
    def hours(self):
        return (self.data['tend'] - self.data['tstart']).sum() / 3600
    
    def __repr__(self):
        return f"Dataset: {len(self)} batches x {self.bs} samples, {self.hours():.1f} hours)"
    
    def get_subset(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None: return self.files
        return self.files.iloc[worker_info.id::worker_info.num_workers]

    def chunk_iterator(self, fname):
        audio = self.load_audio(fname)
        data = self.data[self.data['fname'] == fname]
        for i,row in data.iterrows():
            ts,te = row['tstart'], row['tend']
            sr = self.sample_rate
            chunk = audio[0,int(ts*sr):int(te*sr)]
            if 'ttoks' in row:
                yield chunk,row['ttoks'].tolist()
            else:
                yield chunk,self.tokenizer.encode(row['text'])
    
    def transform_audio(self, chunk):
        sr = self.sample_rate
        padding = self.audio_seconds*sr-chunk.shape[-1]
        lpad = random.randint(0, padding) if self.random_shift else 0
        # masks (downsampled to the Whisper encoder token rate of 50/s) are used
        # to teach the model the concept of padding
        # this let's us decode shorter sequences later
        content_mask = F.pad(torch.full_like(chunk, True, dtype=torch.bool), (lpad, padding-lpad))
        return F.pad(chunk, (lpad, padding-lpad)), content_mask[::320]
    
    def transform_ttoks(self, ttoks):
        rpad = self.ttoks_size - len(ttoks)
        input_toks = F.pad(torch.tensor(list(self.tokenizer.sot_sequence) + ttoks), (0, rpad), value=self.tokenizer.eot)
        output_toks = F.pad(torch.tensor(ttoks + [self.tokenizer.eot]), (0, rpad), value=-100)
        return input_toks, output_toks
    
    def __iter__(self):
        files = self.get_subset().to_list()
        random.shuffle(files)
        files_iter = iter(files)
        # try to have buffer_count buffers but still continue working if we don't have enough files available
        buffers = [self.chunk_iterator(fname) for _, fname in zip(range(self.buffer_count), files_iter)]
        while len(buffers) != 0:
            bx = []
            masks = []
            bitoks = []
            botoks = []
            # select samples from buffered files in a round-robin fashion
            random.shuffle(buffers)
            for bi in range(self.bs):
                if len(buffers) == 0:
#                     print('ran out of buffers in the middle of batch')
                    break
                while len(buffers) != 0:
                    bufi = bi % len(buffers)
                    try:
                        # get a new sample from the buffer
                        x,y = next(buffers[bufi])
                        break
                    except StopIteration:
#                         print('buffer ended')
                        # drop the empty buffer
                        buffers.pop(bufi)
                        try:
                            # replace with a new file
                            buffers.append(self.chunk_iterator(next(files_iter)))
                        except StopIteration:
                            # continue with a reduced number of buffers if we run out of files
                            pass
                x, mask = self.transform_audio(x)
                bx.append(x)
                masks.append(mask)
                itoks, otoks = self.transform_ttoks(y)
                bitoks.append(itoks)
                botoks.append(otoks)
            yield torch.stack(bx, dim=0), torch.stack(masks, dim=0), torch.stack(bitoks, dim=0), torch.stack(botoks, dim=0)

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 16
def load_datasets(
        dataset_path:Path,     # basedir
        index_path:Path,       # feather index file
        subsample:float=1,     # use a fraction of the files
        val_split:float=0.001, # how much data to use for validation
    ):
    basedir = Path(dataset_path)
    data = pd.read_feather(index_path)
    
    tokenizer = whisper.tokenizer.get_tokenizer(False, language='en')
    
    data['fname'] = data['name']
    data['tstart'] = data['ts']
    data['tend'] = data['te']

    # drop first and last segments from each file since they seem to have a lot of missing transcript words
    jmax = data.merge(data.groupby('fname')['j'].max().rename('j_max'), on='fname')
    data['ch_end'] = data['j'] == jmax['j_max']
    first = data['j'] == 0
    last = data['ch_end'] == True
    data_cleaned = data[~first & ~last]

    # drop the last, smaller batch
    val_data, train_data = data_cleaned[:308], data_cleaned[308:]
    train_data = train_data[:len(train_data)//14*14]

    return (
        SADataset(basedir, train_data, tokenizer, bs=32, buffer_count=64),
        SADataset(basedir, val_data, tokenizer, bs=32, buffer_count=64)
    )

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 18
from whisperspeech.train import *
from whisperspeech.modules import *

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 19
import dataclasses

def rand(start, end):
    return random.random() * (end - start) + start

@dataclasses.dataclass
class Tunables:
    init_std :float = 9
    embeddings_std :float = 0.2
    embeddings_lr_scale: float = 10
    output_mult :float = 5.6
    query_mult :float = .3
    rope :bool = True
    mask_embs :bool = True # force embeddings corresponding to the input audio padding to a constant value
    downsample_conv: bool = False
    downsample_mean: bool = True
    
    lr0 :float = 3e-3
    clip_gradient_norm :float = 2
    weight_decay :float = 1e-3
    warmup_steps :float = 2000

    random :bool = False

    def __post_init__(self):
        # randomize the hyperparams if requested
        if self.random:
            self.init_std = 2*10**rand(0,1)
#             self.init_std = 10**rand(-2,1.3)
            self.embeddings_std = 10**rand(-1.7,-0.22)
#             self.embeddings_std = 10**rand(-2,.5)
            self.embeddings_lr_scale = 2**rand(2,4)
#             self.embeddings_lr_scale = rand(5,10)
            self.output_mult = 2**rand(1.5,3)
#             self.output_mult = 2**rand(-1,1)
#             self.output_mult = rand(0.25,0.65)
            self.query_mult = 2**rand(-3,-1.3)
            self.rope = True
            
#             self.lr0 = 10**rand(-3,-2)
#             self.lr0 = rand(1,5)*1e-3
            self.lr0 = 3e-3
            self.clip_gradient_norm = 10**rand(-1,1)
            self.warmup_steps = 100*(10**rand(1.18,1.3))
            
    @staticmethod
    def upgrade(args):
        args = {k:v for k,v in args.items()}
        def old_default(name, value):
            if name not in args: args[name] = value
        old_default('output_mult', 1)
        old_default('query_mult', 1)
        old_default('rope', False)
        old_default('mask_embs', False)
        old_default('downsample_conv', False)
        old_default('downsample_mean', False)
        if 'encoder_depth_ratio' in args: del args['encoder_depth_ratio']
        return args

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 20
import math

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 21
class RQBottleneckTransformer(nn.Module):
    def __init__(self, vq_codes=512, q_depth=12, depth=1, n_head=2, head_width=64, ffn_mult=4,
                 codebook_dim=2, threshold_ema_dead_code=2, use_cosine_sim = False, kl_loss_mul=1,
                 downsample=1,
                 whisper_model_name='tiny.en', tunables=Tunables()):
        super().__init__()
        width = n_head * head_width
        store_attr("codebook_dim,vq_codes,q_depth,n_head,head_width,ffn_mult,depth,use_cosine_sim,downsample,whisper_model_name")
        self.width = width
        self.base_width = 3 * head_width
        self.vq_codes = vq_codes
        self.tunables = tunables
        self.stoks_len = 1500//downsample
        self.stoks_per_sec = self.stoks_len//30
        
        qk_scale = self.tunables.query_mult * 8 / math.sqrt(head_width)
        
        self.kl_loss_mul = kl_loss_mul
        
        n_mlp = width * ffn_mult
        self.mlp = nn.Sequential(
            nn.Linear(width, n_mlp), nn.GELU(), nn.Linear(n_mlp, width)
        )
        self.mlp_ln = LayerNorm(width)

        if tunables.downsample_conv:
            self.downsample_conv = nn.Conv1d(width, width, kernel_size=3, stride=downsample, padding=1)
        else:
            self.downsample_conv = None
        
        if tunables.mask_embs: vq_codes = vq_codes + 1
        self.rq = ResidualVQ(
            dim = width,
            codebook_size = vq_codes, # codebook size
            decay = 0.8,              # the exponential moving average decay, lower means the dictionary will change faster
            commitment_weight = 1.,   # the weight on the commitment loss
            threshold_ema_dead_code = threshold_ema_dead_code,
            use_cosine_sim = use_cosine_sim,
            codebook_dim = codebook_dim,
            num_quantizers= 1,
        )
        
        self.ce_lossf = nn.CrossEntropyLoss(ignore_index=-100)
        self.kl_lossf = nn.KLDivLoss(reduction='batchmean')

#         self.register_buffer("positional_embedding", sinusoids(1500, width))
        self.positional_embedding = nn.Embedding(1500, width) # FIXME: should be self.stoks_len
#         self.register_buffer("embs_padding", None)
        
        self.out_blocks = nn.Sequential(*[
            ResidualAttentionBlock(width, n_head, qk_scale=qk_scale, ffn_mult=ffn_mult, rope=tunables.rope) for _ in range(depth)
        ])
        self.ln_post = LayerNorm(width)
        
        self.whmodel = None
        
        self.apply(self.init_transformer)
        self.reset_stats()
    
    def init_transformer(self, m):
        if isinstance(m, LinearHead):
            m.no_weight_decay = True
            torch.nn.init.constant_(m.weight, 0)
        elif isinstance(m, QueryHead):
            m.lr_scale = 1/(m.weight.shape[1] / self.base_width)
            torch.nn.init.constant_(m.weight, 0)
        elif isinstance(m, nn.Embedding):
            m.no_weight_decay = True
            m.lr_scale = self.tunables.embeddings_lr_scale
            std = self.tunables.embeddings_std
            torch.nn.init.trunc_normal_(m.weight, std=std, a=-3*std, b=3*std)
        elif isinstance(m, nn.Linear):
            m.lr_scale = 1/(m.weight.shape[1] / self.base_width)
            std = self.tunables.init_std / m.weight.shape[1]
            torch.nn.init.trunc_normal_(m.weight, std=std, a=-3*std, b=3*std)
            if m.bias is not None:
                torch.nn.init.trunc_normal_(m.bias, std=std, a=-3*std, b=3*std)
        elif isinstance(m, nn.LayerNorm):
            m.no_weight_decay = True
            torch.nn.init.constant_(m.bias, 0)
            torch.nn.init.constant_(m.weight, 1)

    #
    # training
    #
    @torch.no_grad()
    def extract_teacher(self, samples, input_toks, output_toks):
        embs = self.whmodel[0].encoder(whisper.log_mel_spectrogram(samples))
        teacher_logits = self.whmodel[0].decoder(input_toks, embs)
        # set teacher logits to 0 for padding positions so KLDivLoss ignores them
        teacher_logits[output_toks == -100] = 0
        return embs, teacher_logits
    
    def downsample_embeddings(self, x):
        if self.downsample_conv is not None:
            return x[:,::self.downsample] + self.downsample_conv(x.transpose(-1,-2)).transpose(-2,-1)
        elif self.tunables.downsample_mean:
            bs,slen,depth = x.shape
            return x.reshape(bs,slen//self.downsample,self.downsample,depth).mean(-2)
        else:
            return x[:,::self.downsample]
    
    def forward(self, samples, mask, input_toks, output_toks):
        embs, teacher_logits = self.extract_teacher(samples, input_toks, output_toks)
        
        x = self.downsample_embeddings(embs)
        x = x + self.mlp(self.mlp_ln(x))
        # VQ bottleneck
        quantized, self.indices, self.commit_loss = self.rq(x)
        self.commit_loss = self.commit_loss.mean()

        x = quantized.repeat_interleave(self.downsample, -2)
        if self.tunables.mask_embs: x[~mask] = self.rq.layers[0].project_out(self.rq.layers[0]._codebook.embed[0,self.vq_codes])
        positions = torch.arange(0, x.shape[-2], dtype=torch.long, device=x.device)
        x = x + self.positional_embedding(positions)
        x = self.ln_post(self.out_blocks(x))
        
        logits = self.whmodel[0].decoder(input_toks, x)
        self.ce_loss = self.ce_lossf(logits.view(-1,logits.shape[-1]), output_toks.view(-1))
        self.kl_loss = self.kl_lossf(F.log_softmax(logits, dim=-1), F.softmax(teacher_logits, dim=-1))
        loss = self.ce_loss + self.kl_loss_mul * self.kl_loss + self.commit_loss
        
        if not self.training:
            valid_toks = output_toks != -100
            self.val_true += (logits.argmax(-1)[valid_toks] == output_toks[valid_toks]).float().sum()
            self.val_total += valid_toks.float().sum()

        return x, loss
                
    def reset_stats(self):
        self.register_buffer('val_true', torch.zeros(1).cuda())
        self.register_buffer('val_total', torch.zeros(1).cuda())

    #
    # inference
    #
    @classmethod
    def load_model(cls, repo_id="collabora/spear-tts-pytorch", filename="whisper-vq-stoks.model", local_filename=None):
        if not local_filename:
            local_filename = hf_hub_download(repo_id=repo_id, filename=filename)
        spec = torch.load(local_filename) 
        vqmodel = cls(**spec['config'], tunables=Tunables(**Tunables.upgrade(spec.get('tunables', {}))))
        vqmodel.load_state_dict(spec['state_dict'])
        vqmodel.eval()
        return vqmodel
    
    def save_model(self, fname, store_parameters=True):
        torch.save(dict(config = self.__stored_args__,
                        tunables = dataclasses.asdict(self.tunables),
                        state_dict = self.state_dict() if store_parameters else None), fname)
        
    def ensure_whisper(self):
        # the list wrapper is a hack to make sure the whole of Whisper is not sucked into self.parameters()
        if self.whmodel is None: self.whmodel = [whisper.load_model(self.whisper_model_name).eval()]
        assert self.whisper_model_name.endswith('.en'), "multilingual models are not supported right now"
        self.decoding_options = whisper.DecodingOptions(language='en')
        self.tokenizer = whisper.tokenizer.get_tokenizer(False, language='en')
    
    def quantize(self, embs):
        x = self.downsample_embeddings(embs)
        x = x + self.mlp(self.mlp_ln(x))
        _, stoks, _ = self.rq(x)
        if self.q_depth == 1:
            stoks = stoks.squeeze(-1)
        return stoks

    def dequantize(self, stoks):
        assert self.q_depth == 1
        assert len(stoks.shape) == 1, "batch processing is not supported"
        if isinstance(stoks, np.ndarray): stoks = torch.tensor(stoks)
        # remove padding
        padding = torch.nonzero(stoks == self.vq_codes)
        if padding.any(): stoks = stoks[:padding[0,0]]
        stoks = F.pad(stoks, (0,self.stoks_len - stoks.shape[-1]), value=self.vq_codes if self.tunables.mask_embs else 0)
        x = self.rq.layers[0]._codebook.embed[0,stoks.to(torch.long).view(-1)]
        x = x.repeat_interleave(self.downsample, -2)
        x = self.rq.layers[0].project_out(x).unsqueeze(0)
        positions = torch.arange(0, x.shape[-2], dtype=torch.long, device=x.device)
        x = x + self.positional_embedding(positions)
        return self.ln_post(self.out_blocks(x))

    def encode_audio(self, audio):
        if isinstance(audio, str):
            x, sr = torchaudio.load(audio)
            x = torchaudio.transforms.Resample(sr, 16000)(x)[0]
            audio = x.unsqueeze(0)
        return self.encode_mel(whisper.log_mel_spectrogram(audio))
    
    def encode_mel(self, mel):
        assert len(mel.shape) == 3, "invalid mel spectrogram shape, expect (batch,chn,time)"
        self.ensure_whisper()
        n = mel.shape[-1]
        if n > whisper.audio.N_FRAMES:
            padding = 0
            padded = mel[:,:,:whisper.audio.N_FRAMES]
            mask = None
        else:
            padding = -n % whisper.audio.N_FRAMES
            padded = F.pad(mel, (0, padding), value=-1.5)
        embs = self.whmodel[0].encoder(padded.to(self.whmodel[0].device))#[:,:n//2]
        stoks = self.quantize(embs)
        if self.tunables.mask_embs:
            return stoks[:,:n//2//self.downsample]
        else:
            return stoks
    
    def decode_text(self, stoks, decoding_options=None):
        self.ensure_whisper()
        if decoding_options is None: decoding_options = self.decoding_options
        embs = self.dequantize(stoks).to(self.whmodel[0].device)
        return self.whmodel[0].decode(embs, decoding_options)

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 22
import pylab as plt
import fastprogress
import IPython
import numpy as np

class RQVisual:
    def __init__ (self, model, masterbar, total_steps):
        self.model = model
        self.masterbar = masterbar
        self.total_steps = total_steps
        self.epochs = total_steps // masterbar.main_bar.total
        
        gs = plt.GridSpec(3, 1, height_ratios=[2,2,1])
        graph_fig = plt.figure(figsize=(10,6))
        self.graph_fig = graph_fig
        self.loss_p = graph_fig.add_subplot(gs[0])
        self.acc_p = graph_fig.add_subplot(gs[1], sharex=self.loss_p)
        self.acc_p.tick_params('x', labelbottom=False)
        self.acc_2 = self.acc_p.twinx()
        self.lr_p = graph_fig.add_subplot(gs[2], sharex=self.loss_p)
        self.lr_p.tick_params('x', labelbottom=False)
        self.graph_out = None
        
        self.its = []
        self.train_losses = []
        self.val_losses = []
        self.lr_history = []
        self.entropy_history = []
        self.acc_history = []
            
    def show(self):
        self.start_t = time.time()
        self.masterbar.write(["samples", "train", "val", "codebook entropy", "acc", "time"], table=True)
        self.graph_out = display(self.graph_fig, display_id=True)
        self.entropy_out = display(IPython.display.HTML(''), display_id=True)
    
    def hide(self):
        if self.graph_out is not None:
            self.graph_out.update(IPython.display.HTML(''))
    
    def plot(self):
        loss_p, acc_p, acc_2, lr_p = self.loss_p, self.acc_p, self.acc_2, self.lr_p
        loss_p.clear()
        loss_p.plot(self.its, self.train_losses)
        loss_p.plot(self.its, self.val_losses)
        loss_p.set_xlim(10000, self.total_steps)
        loss_p.set_xscale('log')
        acc_p.clear()
        acc_p.plot(self.its, np.stack(self.entropy_history), ':')
        acc_2.clear()
        acc_2.plot(self.its, np.stack(self.acc_history))
#         acc_p.plot(self.its, np.stack(self.pacc_history), label=range(len(self.pacc_history[0])))
        lr_p.clear()
        lrs = np.array(self.lr_history)
        lr_p.plot(self.its, lrs)
        self.graph_out.update(self.graph_fig)
    
    def add_data(self, it, lr, train_loss, val_los):
        self.its.append(it)
        self.train_losses.append(train_loss)
        self.val_losses.append(val_los)
        self.lr_history.append(lr)
        with torch.no_grad():
            cls = vqmodel.rq.layers[0]._codebook.cluster_size
            pdf = cls / cls.sum()
            entropy = -torch.nansum(pdf * pdf.log2())
            self.acc_history.append((self.model.val_true / self.model.val_total * 100)[0].cpu().numpy())
            self.entropy_history.append(entropy.cpu().numpy())
        if self.entropy_history and self.acc_history:
            self.entropy_out.update(IPython.display.Markdown(f"**Entropy:** {self.entropy_history[-1]:.2f}\n  **Accuracy:** {self.acc_history[-1]:.1f}%"))
        self.model.reset_stats()
        self.plot()

    def add_table_row(self, it, avg_train_loss, val_loss):
        elapsed_t = time.time() - self.start_t
        self.masterbar.write([it, f"{avg_train_loss:.5f}", f"{val_loss:.5f}", f"{self.entropy_history[-1]:.2f}", f"{self.acc_history[-1]:.2f}%", fastprogress.core.format_time(elapsed_t)], table=True)
    
    def on_iter(self, bar, it, avg_train_loss, val_loss):
        epoch = math.ceil(it / self.total_steps * self.epochs)
        bar.comment = f"#{epoch}/{self.epochs} loss: {avg_train_loss:.3f} / {val_loss:.3f}"

# %% ../nbs/2A. VQ semantic token extraction model-Copy1.ipynb 23
def make_model(size:str, tunables:Tunables=Tunables(), dataset:torch.utils.data.Dataset=None):
    if size == 'base.en-2d-4096c':
        model = RQBottleneckTransformer(codebook_dim=32, vq_codes=4096, q_depth=1, n_head=8, depth=1,
                                        downsample=2, threshold_ema_dead_code=0, use_cosine_sim=True,
                                        whisper_model_name=size.split("-")[0], tunables=tunables)
        model.ensure_whisper()
        return model
    raise ArgumentError(f"invalid model size: {size}")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb.

# %% auto 0
__all__ = ['CharTokenizer', 'load_datasets', 'rand', 'Tunables', 'Encoder', 'Decoder', 'TSARTransformer', 'make_model']

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 1
import dataclasses
import random
import math
import torch
import torch.nn as nn
from torch.profiler import record_function

from huggingface_hub import hf_hub_download
from fastcore.basics import store_attr
from fastprogress import progress_bar

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 2
from pathlib import Path
import pylab as plt
import pandas as pd
import numpy as np

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 3
import whisper
from whisperspeech.train import *
from whisperspeech.modules import *

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 10
import torch.nn.functional as F

class SADataset(torch.utils.data.Dataset):
    def __init__(self, data, tokenizer, stoks_config, ttoks_len=550):
        self.data = data
        self.tokenizer = tokenizer
        self.stoks_len = 1500//(stoks_config['config']['downsample'] if stoks_config else 1)
        self.stoks_codes = stoks_config['config']['vq_codes'] if stoks_config else 1024
        if stoks_config['tunables']['mask_embs']: self.stoks_codes += 1
        self.ttoks_len = ttoks_len
    
    def __len__(self):
        return len(self.data)
            
    def __repr__(self):
        return f"<Dataset: {len(self)} samples>"
    
    def __getitem__(self, idx):
        row = self.data.iloc[idx]
        t = row['tend'] - row['tstart']
        Stoks = row['stoks']
        Ttoks = list(row['txt_large'].strip())
        Ttoks = self.tokenizer.encode(row['txt_large'])
        return F.pad(torch.tensor(Ttoks), (0, self.ttoks_len - len(Ttoks)), value=self.tokenizer.eot).to(torch.long), \
               F.pad(torch.tensor(Stoks), (0, self.stoks_len - len(Stoks)), value=self.stoks_codes-1).to(torch.long)

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 11
import re

class CharTokenizer:
    eot = 0
    
    def encode(self, txt):
        return list(bytes(txt.strip(), 'utf-8'))

    def decode(self, tokens):
        return bytes(tokens).decode('utf-8')

def load_datasets(txt_path:Path, stoks_path:Path):
#     tokenizer = whisper.tokenizer.get_tokenizer(multilingual=True)
    tokenizer = CharTokenizer()
    txt_data = pd.read_feather(txt_path)
    stoks_data = pd.read_feather(stoks_path)
    try: stoks_config = torch.load(Path(stoks_path).with_suffix('.model-config'))
    except FileNotFoundError: stoks_config = None
    data = txt_data.merge(stoks_data, on=['fname','j'], suffixes=["", "_2"])
    
    global data_clean
    # drop first and last segments from each file since they seem to have a lot of missing transcript words
    jmax = data.merge(data.groupby('fname')['j'].max().rename('j_max'), on='fname')
    data['ch_end'] = data['j'] == jmax['j_max']
    first = data['j'] == 0
    last = data['ch_end'] == True
    data_clean = data[~first & ~last & (data['txt_large'].str.len() < 550)]
#     data['txt_tiny_wc'] = data['txt_tiny'].map(lambda x: len(re.findall(r'\w+', x)))
#     data['txt_large_wc'] = data['txt_large'].map(lambda x: len(re.findall(r'\w+', x)))
#     data['wc_err'] = data['txt_large_wc'] - data['txt_tiny_wc']
#     data_clean = data[(data['wc_err'].abs() < 5) & (data['txt_large'].str.len() < 550)]
    
    val_data, train_data = data_clean[:300], data_clean[300:]

    return SADataset(train_data, tokenizer, stoks_config), SADataset(val_data, tokenizer, stoks_config)

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 20
def rand(start, end):
    return random.random() * (end - start) + start

@dataclasses.dataclass
class Tunables:
    init_std :float = 1
    embeddings_std :float = .01
    embeddings_lr_scale: float = 5
    output_mult :float = .35
    query_mult :float = 1
    encoder_depth_ratio :float = 0.25
    eot_dropout_p :float = .5
        
    lr0 :float = 3e-3
    clip_gradient_norm :float = .2
    weight_decay :float = 1e-1
    warmup_steps :float = 4000

    random :bool = False

    def __post_init__(self):
        # randomize the hyperparams if requested
        if self.random:
#             self.init_std = 10**rand(-3,1)
            self.init_std = 10**rand(-1,1)
#             self.embeddings_std = 10**rand(-3,1)
            self.embeddings_std = 10**rand(-3,-.7)
#             self.embeddings_lr_scale = 2**rand(-3,3)
            self.embeddings_lr_scale = rand(2,6) #2**rand(0,3)
#             self.output_mult = 2**rand(-2,3)
            self.output_mult = rand(0.25,0.65)
            self.query_mult = 2**rand(-2,3)
#             self.encoder_depth_ratio = random.choice([0.25,0.5,0.75])
            self.encoder_depth_ratio = 0.25
            
#             self.lr0 = 10**rand(-4,-1.5)
            self.lr0 = rand(1,5)*1e-3
            self.clip_gradient_norm = 10**rand(-3,0)
            self.warmup_steps = 100*(10**rand(1,1.85))

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 21
class Encoder(nn.Module):
    def __init__(self, depth=6, width=384, n_head=6, length=1500, codes=1024, emb_width=384, ffn_mult=4, pos_embs=None, tunables=Tunables()):
        super().__init__()
        self.emb_width = emb_width
        
        self.emb_factor = width != emb_width
    
        self.embedding = nn.Embedding(codes, emb_width)
        if self.emb_factor:
            self.emb_to_hidden = nn.Linear(emb_width, width)

        if pos_embs is None: pos_embs = sinusoids(length, width)
        self.register_buffer("positional_embedding", pos_embs)

        self.layers = nn.Sequential(*[
            ResidualAttentionBlock(width, n_head,
                                   qk_scale=tunables.query_mult*8/math.sqrt(width/n_head), ffn_mult=ffn_mult) for _ in range(depth)
        ])

        self.ln_post = LayerNorm(width)
        
    def forward(self, Stoks):
        xin = self.embedding(Stoks)
        if self.emb_factor:
            xin = self.emb_to_hidden(xin)
        
        assert xin.shape[1:] == self.positional_embedding.shape, "incorrect semantic token shape"
        xin = (xin + self.positional_embedding).to(xin.dtype)

        return self.ln_post(self.layers(xin))

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 22
class Decoder(nn.Module):
    def __init__(self, depth=6, stoks_width=384, width=384, n_head=6, length=1500, codes=1024, ffn_mult=4, pos_embs=None, tunables=Tunables()):
        super().__init__()
        self.length = length
        self.codes = codes
        self.width = width
        self.stoks_width = stoks_width
        
        self.emb_factor = width != stoks_width
    
        # embed semantic tokens
        self.embedding = nn.Embedding(codes, stoks_width)
        if self.emb_factor:
            self.emb_to_hidden = nn.Linear(stoks_width, width)
            self.hidden_to_emb = nn.Linear(width, stoks_width)

        if pos_embs is None: pos_embs = sinusoids(length, width)
        self.register_buffer("positional_embedding", pos_embs)
        
        self.layers = nn.ModuleList([
            ResidualAttentionBlock(width, n_head, cross_attention=True,
                                   qk_scale=tunables.query_mult*8/math.sqrt(width/n_head), ffn_mult=ffn_mult) for _ in range(depth)
        ])
        self.ln_post = LayerNorm(width)
        
    def forward(self, Stoks, xenc):
        sot = self.embedding(torch.tensor([self.codes-1]).to(Stoks.device)).repeat(Stoks.shape[0],1,1)
        if Stoks.shape[-1] > 0:
            if Stoks.shape[-1] >= self.length:
                Stoks = Stoks[:,:-1]
            Sembs = self.embedding(Stoks)
            Sembs = torch.cat([sot, Sembs], dim=-2)
        else:
            Sembs = sot
            
        if self.emb_factor:
            Sembs = self.emb_to_hidden(Sembs)
    
        xin = (Sembs + self.positional_embedding[:Sembs.shape[1]]).to(xenc.dtype)
    
        x = xin
        for l in self.layers: x = l(x, xenc, causal=True)
        
        x = self.ln_post(x)
    
        if self.emb_factor:
            x = self.hidden_to_emb(x)
        
        logits = (x @ self.embedding.weight.to(x.dtype).T).float()
        return logits

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 23
class TSARTransformer(nn.Module):
    def __init__(self, depth=6, n_head=6, head_width=64, ffn_mult=4, language='en',
                 ttoks_len=200, ttoks_codes=50364, ttoks_width=None,
                 stoks_len=1500, stoks_codes=1024, stoks_width=None,
                 tunables=Tunables()):
        assert language == 'en', "only english is supported right now"
        super().__init__()
        store_attr("depth,n_head,head_width,ffn_mult,stoks_width,ttoks_width,ttoks_len,stoks_len,ttoks_codes,stoks_codes,language")

        width = n_head * head_width
        self.width = width
        self.base_width = 3 * head_width
        self.tunables = tunables
        if self.stoks_width is None: self.stoks_width = self.width
        if self.ttoks_width is None: self.ttoks_width = self.width
        
        encoder_depth = int(depth * 2 * tunables.encoder_depth_ratio)
        decoder_depth = depth * 2 - encoder_depth
        tformer_args = dict(width=width, n_head=n_head, ffn_mult=ffn_mult, tunables=tunables)
        self.encoder = Encoder(length=ttoks_len, codes=ttoks_codes, emb_width=self.ttoks_width, depth=encoder_depth, **tformer_args)
        self.decoder = Decoder(length=stoks_len, codes=stoks_codes, stoks_width=self.stoks_width, depth=decoder_depth, **tformer_args)
        
        self.tokenizer = None
        
        self.apply(self.init_transformer)

    def init_transformer(self, m):
        if isinstance(m, LinearHead):
            m.no_weight_decay = True
            torch.nn.init.constant_(m.weight, 0)
        elif isinstance(m, QueryHead):
            m.lr_scale = 1/(m.weight.shape[1] / self.base_width)
            torch.nn.init.constant_(m.weight, 0)
        elif isinstance(m, nn.Embedding):
            m.no_weight_decay = True
            m.lr_scale = self.tunables.embeddings_lr_scale
            std = self.tunables.embeddings_std
            torch.nn.init.trunc_normal_(m.weight, std=std, a=-3*std, b=3*std)
        elif isinstance(m, nn.Linear):
            m.lr_scale = 1/(m.weight.shape[1] / self.base_width)
            std = self.tunables.init_std / m.weight.shape[1]
            torch.nn.init.trunc_normal_(m.weight, std=std, a=-3*std, b=3*std)
            if m.bias is not None:
                torch.nn.init.trunc_normal_(m.bias, std=std, a=-3*std, b=3*std)
        elif isinstance(m, nn.LayerNorm):
            m.no_weight_decay = True
            torch.nn.init.constant_(m.bias, 0)
            torch.nn.init.constant_(m.weight, 1)
        
    def forward(self, Ttoks, Stoks, loss=True):
        with record_function("encoder"):
            xenc = self.encoder(Ttoks.to(torch.long))
        with record_function("decoder"):
            if self.training and self.tunables.eot_dropout_p > 0:
                is_eot = Stoks == self.stoks_codes - 1
                is_eot[torch.rand(Stoks.shape[0]) < self.tunables.eot_dropout_p] = 0
                Stoks_rand = torch.randint_like(Stoks, self.stoks_codes)
                Stoks_ = torch.where(is_eot, Stoks_rand, Stoks)
            else:
                Stoks_ = Stoks
            logits = self.decoder(Stoks, xenc) * self.tunables.output_mult / (self.width / self.base_width)
        if loss is not None:
            with record_function("loss"):
                loss = F.cross_entropy(logits.transpose(-1,-2), Stoks, reduction='none')
        return logits, loss

    #
    # inference
    #
    @classmethod
    def load_model(cls, repo_id="collabora/whisperspeech", filename="t2s_up.model", local_filename=None):
        if not local_filename:
            local_filename = hf_hub_download(repo_id=repo_id, filename=filename)
        spec = torch.load(local_filename)
        model = cls(**spec['config'], tunables=Tunables(**spec['tunables']))
        model.load_state_dict(spec['state_dict'])
        model.eval()
        return model

    def load_checkpoint(self, local_filename):
        spec = torch.load(local_filename, map_location='cpu')
        assert 'pytorch-lightning_version' in spec, 'not a valid PyTorch Lightning checkpoint'
        state_dict = {k.replace('model.', ''):v
                      for k,v in spec['state_dict'].items()}
        self.load_state_dict(state_dict)
        return self

    def save_model(self, fname):
        torch.save(dict(config = self.__stored_args__,
                        tunables = dataclasses.asdict(self.tunables),
                        state_dict = self.state_dict()), fname)

    def ensure_tokenizer(self):
        assert not self.training
        if self.tokenizer is None: self.tokenizer = CharTokenizer()
        #whisper.tokenizer.get_tokenizer(multilingual=True)  

    @property
    def device(self):
        return next(self.parameters()).device
    
    @torch.no_grad()
    def generate(self, txt, N=None, T=0.7, top_k=None, show_progress_bar=True):
        self.ensure_tokenizer()
        N = N or self.stoks_len
        dev = self.device
        ttoks = torch.tensor(self.tokenizer.encode(txt), device=dev)
        ttoks = F.pad(ttoks, (0, self.ttoks_len - len(ttoks)), value=self.tokenizer.eot).unsqueeze(0)
        toks = torch.zeros((1,N), dtype=torch.long, device=dev)
        it = range(N)
        if show_progress_bar: it = progress_bar(it)
        for i in it:
            p, _ = self(ttoks, toks[:,:i], loss=None)
            last_p = p[0,-1]
            if top_k:
                last_p[last_p < torch.topk(last_p, top_k).values[-1,None]] = -torch.inf
            tok = torch.multinomial((last_p / float(T)).softmax(-1), 1)
            toks[0,i] = tok
            if toks[0,i] == self.stoks_codes-1: return toks[0,:i]
        return toks[0]
    
    @torch.no_grad()
    def generate_batch(self, txts, N=None, T=1.1, top_k=7, show_progress_bar=True):
        self.ensure_tokenizer()
        N = self.stoks_len
        dev = self.device
        ttoks = []
        for txt in txts:
            ttoks_ = torch.tensor(self.tokenizer.encode(txt), device=dev)
            ttoks_ = F.pad(ttoks_, (0, self.ttoks_len - len(ttoks_)), value=self.tokenizer.eot).unsqueeze(0)
            ttoks.append(ttoks_)
        ttoks = torch.cat(ttoks, dim=0)
        toks = torch.zeros((len(ttoks),N), dtype=torch.long, device=dev)
        it = range(N)
        if show_progress_bar: it = progress_bar(it)
        for i in it:
            p, _ = self(ttoks, toks[:,:i], loss=None)
            last_p = p[:,-1]
            if top_k:
                last_p[last_p < torch.topk(last_p, top_k).values[:,-1,None]] = -torch.inf
            tok = torch.multinomial((last_p / float(T)).softmax(-1), 1)
            toks[:,i] = tok[:,0]
            if (toks[:,i] == self.stoks_codes-1).all(): return toks[:,:i]
        return toks

# %% ../nbs/5. Text to semantic token modeling μP graphemes cleaned.ipynb 24
def make_model(size:str, tunables:Tunables=Tunables(), dataset:SADataset=None, **kwargs):
    kwargs = dict(stoks_len = dataset.stoks_len, stoks_codes=dataset.stoks_codes, ttoks_len = dataset.ttoks_len, tunables=tunables, **kwargs)
    if size == 'micro':
        return TSARTransformer(depth=2, n_head=3, ffn_mult=1, **kwargs)
    if size == 'tiny':
        return TSARTransformer(depth=4, n_head=6, **kwargs)
    if size == 'base':
        return TSARTransformer(depth=6, n_head=8, **kwargs)
    if size == 'small':
        return TSARTransformer(depth=12, n_head=16, **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/1B. Voice activity detection.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/1B. Voice activity detection.ipynb 3
import os
import random
import torch
import torchaudio

from pathlib import Path
from fastprogress import progress_bar
from fastcore.script import call_parse

import numpy as np
import webdataset as wds
from huggingface_hub import hf_hub_download

import whisperx
from speechbrain.pretrained import EncoderClassifier
from clara import PLCLARA
import clara

# %% ../nbs/1B. Voice activity detection.ipynb 5
# some of the original file names have a dot in their name
# webdataset does not like it so let's patch it
def fix_dots_in_names(name):
    name, ext = name.rsplit('.', 1)
    return ".".join((name.replace('.', '_'), ext))

def load_dataset(url, decode=True, rename_files=None):
    ds = wds.WebDataset(url, rename_files=rename_files)
    if not decode: return ds
    return ds.decode(wds.torch_audio).compose(
        wds.rename(audio="flac;mp3;wav;ogg")
    )

# %% ../nbs/1B. Voice activity detection.ipynb 7
def extract_segments(vad_result, max_duration):
    binarize = whisperx.vad.Binarize(max_duration=max_duration)
    segments = binarize(vad_result)
    return [(x.start, x.end) for x in segments.get_timeline()]

def segment_audio(vad_model, audio, sr=16000):
    vad_result = vad_model({"waveform": audio, "sample_rate": sr})
    return extract_segments(vad_result, 30)

# %% ../nbs/1B. Voice activity detection.ipynb 25
def split_to_chunks(stream, pad_to_seconds=None):
    for s in stream:
        audio, sr = s['audio']
        imax = len(s['vad.npy']) - 1
        for i,(ts,te) in enumerate(s['vad.npy']):
            samples = audio[0,int(ts*sr):int(te*sr)]
            if pad_to_seconds is not None:
                padding = pad_to_seconds*sr-samples.shape[-1]
                lpad = random.randint(0, padding) if random_shift else 0
                samples = F.pad(samples, (lpad, padding-lpad))
            yield {"__key__": s['__key__'] + f"_{i:03d}",
                   "__url__": s['__url__'],
                   "i": i, "imax": imax,
                   "tstart": ts, "tend": te, "total_seconds": audio.shape[-1]/sr,
                   "lpad": lpad, "rpad": padding-lpad,
                   "lpad_s": lpad/sr, "rpad_s": (padding-lpad)/sr,
                   "samples": samples, "sample_rate": sr}

# %% ../nbs/1B. Voice activity detection.ipynb 26
def flac_to_vad_name(input, key='raw'):
    if '-flac-' in input:
        return input.rsplit("/", 1)[1].replace('flac', 'vad') + ".gz"
    else:
        return input.rsplit("/", 1)[1].replace(key, 'vad') + ".gz"

@call_parse
def process_shard(
    input:str,           # input shard URL/path
    output:str=None,     # output shard URL/path
    key:str='raw',       # string to replace with 'vad' in the shard name
):
    if output is None: output = flac_to_vad_name(input, key=key)
    
    rename_files=fix_dots_in_names if 'librilight' in input else None
    ds = load_dataset(input, rename_files=rename_files)
    dl = torch.utils.data.DataLoader(ds, num_workers=1, batch_size=None)
    vad_model = whisperx.vad.load_vad_model('cuda')
    
    tmp = output+".tmp"
    last_key = None
    with wds.TarWriter(tmp) as sink:
        for s in progress_bar(dl, total='noinfer'):
            audio, sr = s.get('audio', (None, None))
            if audio is None:
                print(f"warning: '{s['__key__']}' does not contain an audio file")
                continue
            sink.write({
                "__key__": s['__key__'],
                "vad.npy": np.array(segment_audio(vad_model, audio, sr=sr), dtype=np.float32),
            })
            last_key = s['__key__']
    os.rename(tmp, output)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/7. Pipeline.ipynb.

# %% auto 0
__all__ = ['Pipeline']

# %% ../nbs/7. Pipeline.ipynb 1
import torch
from whisperspeech.t2s_up import TSARTransformer
from whisperspeech.s2a_delar_mup import SADelARTransformer
from whisperspeech.a2wav import Vocoder

# %% ../nbs/7. Pipeline.ipynb 2
class Pipeline:
    def __init__(self):
#         self.t2s = TSARTransformer.load_model().cuda()
#         self.s2a = SADelARTransformer.load_model().cuda()
        self.vocoder = Vocoder()

    def generate_atoks(self, text, speaker="default", show_progress_bar=True):
        if speaker == "default":
            speaker = "3645" if "3645" in self.s2a.speaker_map else next(iter(self.s2a.speaker_map.keys()))
        text = text.replace("\n", " ")
        stoks = self.t2s.generate(text, T=.5, top_k=3, show_progress_bar=show_progress_bar)
        atoks = self.s2a.generate(stoks, [speaker], T=2, top_k=8, show_progress_bar=show_progress_bar)
        return atoks
        
    def generate(self, text, speaker="default", show_progress_bar=True):
        return self.vocoder.decode(self.generate_atoks(text, speaker, show_progress_bar=show_progress_bar))
    
    def generate_to_file(self, fname, text, speaker="default", show_progress_bar=True):
        self.vocoder.decode_to_file(fname, self.generate_atoks(text, speaker, show_progress_bar=show_progress_bar))
        
    def generate_to_notebook(self, text, speaker="default", show_progress_bar=True):
        self.vocoder.decode_to_notebook(self.generate_atoks(text, speaker, show_progress_bar=show_progress_bar))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/4C. Semantic to acoustic modeling eval.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/4C. Semantic to acoustic modeling eval.ipynb 1
import io
import time
import math
import random
import dataclasses

# %% ../nbs/4C. Semantic to acoustic modeling eval.ipynb 2
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.profiler import profile, record_function, ProfilerActivity, schedule
from fastcore.basics import store_attr
from huggingface_hub import hf_hub_download

# %% ../nbs/4C. Semantic to acoustic modeling eval.ipynb 3
from pathlib import Path
import json
from fastprogress import progress_bar, master_bar
import webdataset as wds

# %% ../nbs/4C. Semantic to acoustic modeling eval.ipynb 4
from .train import *
from .modules import *
from . import vq_stoks, s2a_delar_mup_wds_2 as s2a_delar
